from __future__ import annotations
from collections import defaultdict


class StudentAllocationNotPossible(Exception): pass
courses_switched_to: dict[Course, Switch] = {}


def students_to_courses(C: dict[str, int], S: dict[str, list]) -> list[int]:
    ''' Solves the problem of assigning n students i=0,1,2,...,n-1 to m courses j=0,1,2,...,m-1, where each students 
        chooses a list of courses. Time/space complexity O(nm). Returns a list s2c with s2c[i] being the course to which
        student i was assinged to.
    
    >>> course_capacities = {'1': 1, '2': 2, '3': 3}
    >>> student_votes = {'1': ['1', '3'], '2': ['1', '3'], '3': ['2', '3'], '4': ['1'], '5': ['1', '2', '3'], '6': ['3']}
    >>> students_to_courses(course_capacities, student_votes)
    {'1': '3', '2': '3', '3': '2', '4': '1', '5': '2', '6': '3'}
    '''    
    courses = {id: Course(id=id, capacity=int(cap)) for id, cap in C.items()}
    students = [Student(id=id, voted_courses=[courses[j] for j in vote]) for id, vote in S.items()]

    for student in students:
        student.assign()

    return {student.id: student.assigned_course.id if student.assigned_course else None for student in students }


class Course:
    __slots__ = ('students', 'id', 'neighbours', 'capacity')
    students: set[Student]
    id: int
    neighbours: defaultdict[Course, set[Student]]
    capacity: int

    def __init__(self, id: str, capacity: int) -> None:
        self.id = id
        self.students = set()
        self.neighbours = defaultdict(set)
        self.capacity = capacity
    
    def __hash__(self) -> int:
        return hash(self.id)

    def __repr__(self) -> str:
        return f'Course(id={self.id}, capacity={self.capacity}, students={len(self.students)})'

    def __eq__(self, __o) -> bool:
        return self.id == __o.id

    def add_student(self, student: Student):
        self.students.add(student)
        for course in student.voted_courses:
            if course != self:
                self.neighbours[course].add(student)
    
    def remove_student(self, student: Student):
        self.students.remove(student)
        for course in student.voted_courses:
            if course != self:
                self.neighbours[course].remove(student)
            if not self.neighbours[course]:
                del self.neighbours[course]

    def is_full(self) -> bool:
        return len(self.students) >= self.capacity


class Switch:
    __slots__ = ('student', 'target', 'source')
    student: Student
    target: Course
    source: Course

    def __init__(self, student: Student, source: Course, target: Course ) -> None:
        self.student = student
        self.source = source
        self.target = target
    
    def __repr__(self) -> str:
        return f'SWITCH {self.student.id}: {self.source.id}->{self.target.id}'


class Student:
    __slots__ = ('id', 'voted_courses', 'assigned_course')
    id: int
    voted_courses: list[Course]
    assigned_course: Course

    def __init__(self, id: str, voted_courses: list[Course]) -> None:
        self.id = id
        self.voted_courses = voted_courses
        self.assigned_course = None
    
    def __hash__(self) -> int:
        return hash(self.id)
    
    def __repr__(self) -> str:
        return f'Student(id={self.id}, voted_courses={list(map(lambda c: c.id, self.voted_courses))})'
    
    def __eq__(self, __o) -> bool:
        return self.id == __o.id

    def __lt__(self, other) -> bool:
        return self.id < other.id

    def assign(self):
        global courses_switched_to
        courses_switched_to = {}

        courses_to_check = {course: self for course in self.voted_courses}
        course = rassign(courses_to_check)
        if course:
            execute_switches(course, self)
        return course


def rassign(courses_to_check: dict[Course, Student]) -> Course:
    global courses_switched_to

    if not courses_to_check:
        return None

    for course, student in courses_to_check.items():
        if course.is_full():
            continue
        else:
            if course in courses_switched_to:
                return course
            if student.assigned_course:
                switch = Switch(student, student.assigned_course, course)
                courses_switched_to[course] = switch
            return course
    else:
        new_courses_to_check = {}
        for course in courses_to_check:
            for neighbour in course.neighbours.keys():
                if neighbour not in courses_switched_to and neighbour not in new_courses_to_check:
                    for s in course.neighbours[neighbour]:
                        student = s
                        break
                    new_courses_to_check[neighbour] = student
                    switch = Switch(student, course, neighbour)
                    courses_switched_to[neighbour] = switch
        return rassign(new_courses_to_check)


def execute_switches(course_to_switch_to: Course, student: Student) -> None:
    if course_to_switch_to in student.voted_courses:
        course_to_switch_to.add_student(student)
        student.assigned_course = course_to_switch_to
    else:
        switch = courses_switched_to[course_to_switch_to]
        switch.source.remove_student(switch.student)
        switch.target.add_student(switch.student)
        switch.student.assigned_course = switch.target
        execute_switches(switch.source, student)
